"""Models for the projectroles app"""

import logging
import re
import uuid

from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractUser, Group
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from djangoplugins.models import Plugin
from markupfield.fields import MarkupField

from projectroles.constants import get_sodar_constants


AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')
logger = logging.getLogger(__name__)


# SODAR constants
SODAR_CONSTANTS = get_sodar_constants()
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_ROLE_FINDER = SODAR_CONSTANTS['PROJECT_ROLE_FINDER']
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']

# Local constants
ROLE_RANKING = {
    PROJECT_ROLE_OWNER: 10,
    PROJECT_ROLE_DELEGATE: 20,
    PROJECT_ROLE_CONTRIBUTOR: 30,
    PROJECT_ROLE_GUEST: 40,
    PROJECT_ROLE_FINDER: 50,
}
PROJECT_TYPE_CHOICES = [('CATEGORY', 'Category'), ('PROJECT', 'Project')]
APP_SETTING_TYPES = ['BOOLEAN', 'INTEGER', 'STRING', 'JSON']
APP_SETTING_TYPE_CHOICES = [
    ('BOOLEAN', 'Boolean'),
    ('INTEGER', 'Integer'),
    ('STRING', 'String'),
    ('JSON', 'Json'),
]
APP_SETTING_VAL_MAXLENGTH = 255
PROJECT_SEARCH_TYPES = ['project']
PROJECT_TAG_STARRED = 'STARRED'
CAT_DELIMITER = ' / '
CAT_DELIMITER_ERROR_MSG = 'String "{}" is not allowed in title'.format(
    CAT_DELIMITER
)
ROLE_PROJECT_TYPE_ERROR_MSG = (
    'Invalid project type "{project_type}" for ' 'role "{role_name}"'
)


# Project ----------------------------------------------------------------------


class ProjectManager(models.Manager):
    """Manager for custom table-level Project queries"""

    def find(self, search_terms, keywords=None, project_type=None):
        """
        Return projects with a partial match in full title or, including titles
        of parent Project objects, or the description of the current object.
        Restrict to project type if project_type is set.

        :param search_terms: Search terms (list)
        :param keywords: Optional search keywords as key/value pairs (dict)
        :param project_type: Project type or None
        :return: QuerySet of Project objects
        """
        projects = super().get_queryset().order_by('title')
        if project_type:
            projects = projects.filter(type=project_type)
        term_query = Q()
        for t in search_terms:
            term_query.add(Q(full_title__icontains=t), Q.OR)
            term_query.add(Q(description__icontains=t), Q.OR)
        return projects.filter(term_query).order_by('full_title')


class Project(models.Model):
    """
    A SODAR project. Can have one parent category in case of nested
    projects. The project must be of a specific type, of which "CATEGORY" and
    "PROJECT" are currently implemented. "CATEGORY" projects are used as
    containers for other projects.
    """

    #: Project title
    title = models.CharField(
        max_length=255, unique=False, help_text='Project title'
    )

    #: Type of project ("CATEGORY", "PROJECT")
    type = models.CharField(
        max_length=64,
        choices=PROJECT_TYPE_CHOICES,
        default=SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'],
        help_text='Type of project ("CATEGORY", "PROJECT")',
    )

    #: Parent category if nested, otherwise null
    parent = models.ForeignKey(
        'self',
        blank=True,
        null=True,
        related_name='children',
        help_text='Parent category if nested',
        on_delete=models.CASCADE,
    )

    #: Short project description
    description = models.CharField(
        max_length=512,
        unique=False,
        blank=True,
        null=True,
        help_text='Short project description',
    )

    #: Project README (optional, supports markdown)
    readme = MarkupField(
        null=True,
        blank=True,
        markup_type='markdown',
        help_text='Project README (optional, supports markdown)',
    )

    #: Public guest access
    public_guest_access = models.BooleanField(
        default=False,
        help_text='Allow public guest access for the project, also including '
        'unauthenticated users if allowed on the site',
    )

    #: Project is archived (read-only)
    archive = models.BooleanField(
        default=False,
        help_text='Project is archived (read-only)',
    )

    #: Full project title with parent path (auto-generated)
    full_title = models.CharField(
        max_length=4096,
        null=True,
        help_text='Full project title with parent path (auto-generated)',
    )

    #: Whether project has children with public access (auto-generated)
    has_public_children = models.BooleanField(
        default=False,
        help_text='Whether project has children with public access '
        '(auto-generated)',
    )

    #: Project SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Project SODAR UUID'
    )

    # Set manager for custom queries
    objects = ProjectManager()

    class Meta:
        unique_together = ('title', 'parent')
        ordering = ['parent__title', 'title']

    def __str__(self):
        return self.full_title

    def __repr__(self):
        values = (
            self.title,
            self.type,
            self.parent.title if self.parent else None,
        )
        return 'Project({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Custom validation and field populating for Project"""
        self._validate_parent()
        self._validate_title()
        self._validate_parent_type()
        self._validate_public_guest_access()
        self._validate_archive()
        # Update full title of self and children
        self.full_title = self._get_full_title()
        if self.type == PROJECT_TYPE_CATEGORY:
            for child in self.children.all():
                child.save()
        # Update public children
        # NOTE: Parents will be updated in ProjectModifyMixin.modify_project()
        self.has_public_children = self._has_public_children()
        super().save(*args, **kwargs)

    def _validate_parent(self):
        """
        Validate parent value to ensure project can't be set as its own parent.
        """
        if self.parent == self:
            raise ValidationError('Project can not be set as its own parent')

    def _validate_parent_type(self):
        """Validate parent value to ensure parent can not be a project"""
        if (
            self.parent
            and self.parent.type == SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
        ):
            raise ValidationError(
                'Subprojects are only allowed within categories'
            )

    def _validate_public_guest_access(self):
        """Validate public guest access to ensure it is not set on categories"""
        if (
            self.public_guest_access
            and self.type == SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
        ):
            raise ValidationError(
                'Public guest access is not allowed for categories'
            )

    def _validate_title(self):
        """
        Validate title against parent title to ensure they don't equal parent.
        """
        if self.parent and self.title == self.parent.title:
            raise ValidationError('Project and parent titles can not be equal')
        if (
            CAT_DELIMITER in self.title
            or self.title.startswith(CAT_DELIMITER.strip())
            or self.title.endswith(CAT_DELIMITER.strip())
        ):
            raise ValidationError(CAT_DELIMITER_ERROR_MSG)

    def _validate_archive(self):
        """
        Validate archive status against project type to ensure archiving is only
        applied to projects.
        """
        if (
            self.archive
            and self.type != SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
        ):
            raise ValidationError(
                'Archiving a category is not currently supported'
            )

    def get_absolute_url(self):
        return reverse(
            'projectroles:detail', kwargs={'project': self.sodar_uuid}
        )

    # Internal helpers

    def _get_full_title(self):
        """Return full title of project with path."""
        parents = self.get_parents()
        ret = (
            CAT_DELIMITER.join([p.title for p in parents]) + CAT_DELIMITER
            if parents
            else ''
        )
        return ret + self.title

    def _has_public_children(self):
        """
        Return True if the project has any children with public guest access.
        """
        if self.type != PROJECT_TYPE_CATEGORY:
            return False
        for child in self.get_children():
            if child.public_guest_access:
                return True
            ret = child._has_public_children()
            if ret:
                return True
        return False

    def _update_public_children(self):
        """Update has_public_children for this project's parents."""
        if self.parent:
            parent = self.parent
            public_found = False
            while parent:
                if public_found:
                    parent.has_public_children = True
                else:
                    parent.has_public_children = parent._has_public_children()
                parent.save()
                if not public_found and parent.has_public_children:
                    public_found = True
                parent = parent.parent

    # Custom row-level functions

    def get_parents(self):
        """
        Return a list of parent projects in inheritance order.

        :return: List of Project objects
        """
        if not self.parent:
            return []
        ret = []
        parent = self.parent
        while parent:
            ret.append(parent)
            parent = parent.parent
        return reversed(ret)

    def get_children(self, flat=False):
        """
        Return child objects for a Category, sorted by full title.

        :param flat: Return all children recursively as a flat list (bool)
        :return: QuerySet
        """
        if self.type != PROJECT_TYPE_CATEGORY:
            return Project.objects.none()
        if flat:
            return Project.objects.filter(
                full_title__startswith=self.full_title + CAT_DELIMITER
            ).order_by('full_title')
        return self.children.all().order_by('title')

    def get_depth(self):
        """Return depth of project in the project tree structure (root=0)"""
        return len(self.full_title.split(CAT_DELIMITER)) - 1

    def get_role(self, user, inherited_only=False):
        """
        Return the currently active role for user, or None if not available.
        Returns the highest ranked role including inherited roles. In case of
        multiple roles of the same level in the hierarchy, the lowest one is
        returned.

        :param user: User object
        :param inherited_only: Only return an inherited role if True
                               (boolean, default=False)
        :return: RoleAssignment object or None
        """
        if not user or user.is_anonymous:
            return None
        projects = [self] if not inherited_only else []
        projects += list(self.get_parents())
        return (
            RoleAssignment.objects.filter(
                project__in=projects,
                user=user,
                role__project_types__contains=[self.type],
            )
            .order_by('role__rank', '-project__full_title')
            .first()
        )

    def get_roles(
        self,
        user=None,
        inherited=True,
        inherited_only=False,
        min_rank=None,
        max_rank=None,
    ):
        """
        Return project role assignments.

        :param user: Limit to user (User object or None)
        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :param min_rank: Limit roles to minimum rank (integer or None)
        :param max_rank: Limit roles to maximum rank (integer or None)
        :return: List of RoleAssignment objects
        :raise: ValueError If inheritance arguments conflict
        """
        if not inherited and inherited_only:
            raise ValueError(
                'Inherited set False and inherited_only set True, No results '
                'can be returned'
            )
        projects = [] if inherited_only else [self]
        # NOTE: We have to get inherited roles to exclude overridden ones
        parents = list(self.get_parents())
        projects += parents
        q_kwargs = {
            'project__in': projects,
            'role__project_types__contains': [self.type],
        }
        if user and user.is_authenticated:
            q_kwargs['user'] = user
        roles = RoleAssignment.objects.filter(**q_kwargs).order_by(
            '-project__full_title', 'role__name', 'user'
        )
        user_roles = {}
        for a in roles:
            u = a.user
            rank_ok = (not min_rank or a.role.rank >= min_rank) and (
                not max_rank or a.role.rank <= max_rank
            )
            # Local role (always returned first if it exists)
            if a.project == self and not inherited_only and rank_ok:
                user_roles[u] = a
            # Inherited role of higher rank
            elif (
                inherited
                and a.project in parents
                and (
                    u not in user_roles or a.role.rank < user_roles[u].role.rank
                )
                and rank_ok
            ):
                user_roles[u] = a
            # Pop overridden role if in list
            elif (
                a.project in parents
                and u in user_roles
                and a.role.rank < user_roles[u].role.rank
            ):
                user_roles.pop(u, None)
        return list(user_roles.values())

    def get_roles_by_rank(
        self, role_name, inherited=True, inherited_only=False
    ):
        """
        Return RoleAssignments for specific role name. Will also include custom
        roles of identical rank once role customization is implemented (see
        issue #288).

        :param role_name: Name of role (string)
        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :return: List
        """
        if role_name not in ROLE_RANKING:
            role = Role.objects.filter(name=role_name).first()
            if not role:
                raise ValueError('Unknown role "{}"'.format(role_name))
            rank = role.rank
        else:
            rank = ROLE_RANKING[role_name]
        return self.get_roles(
            inherited=inherited,
            inherited_only=inherited_only,
            min_rank=rank,
            max_rank=rank,
        )

    def get_owner(self):
        """
        Return RoleAssignment for local (non-inherited) owner or None if not
        set.

        :return: QuerySet
        """
        return self.local_roles.filter(role__name=PROJECT_ROLE_OWNER).first()

    def get_owners(self, inherited=True, inherited_only=False):
        """
        Return RoleAssignments for project owner as well as possible inherited
        owners from parent projects.

        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :return: List
        """
        rank = ROLE_RANKING[PROJECT_ROLE_OWNER]
        return self.get_roles(
            inherited=inherited,
            inherited_only=inherited_only,
            min_rank=rank,
            max_rank=rank,
        )

    def get_delegates(self, inherited=True, inherited_only=False):
        """
        Return RoleAssignments for delegates. Excludes delegates with an
        inherited owner role.

        :param inherited: Include inherited roles (bool)
        :param inherited_only: Return only inherited roles (bool)
        :return: List
        """
        rank = ROLE_RANKING[PROJECT_ROLE_DELEGATE]
        return self.get_roles(
            inherited=inherited,
            inherited_only=inherited_only,
            min_rank=rank,
            max_rank=rank,
        )

    def is_owner(self, user):
        """
        Return True if user is owner in this project or inherits ownership from
        a parent category.

        :return: Boolean
        """
        if not user.is_authenticated:
            return False
        role_as = self.get_role(user)
        if role_as and role_as.role.rank == ROLE_RANKING[PROJECT_ROLE_OWNER]:
            return True
        return False

    def is_delegate(self, user):
        """
        Return True if user is delegate in this project or inherits delegate
        status from a parent category.

        :return: Boolean
        """
        if not user.is_authenticated:
            return False
        role_as = self.get_role(user)
        if role_as and role_as.role.rank == ROLE_RANKING[PROJECT_ROLE_DELEGATE]:
            return True
        return False

    def is_owner_or_delegate(self, user):
        """
        Return True if user is either an owner or a delegate in this project.
        Includes inherited assignments.

        :return: Boolean
        """
        if not user.is_authenticated:
            return False
        role_as = self.get_role(user)
        if role_as and role_as.role.rank in [
            ROLE_RANKING[PROJECT_ROLE_OWNER],
            ROLE_RANKING[PROJECT_ROLE_DELEGATE],
        ]:
            return True
        return False

    def get_members(self, inherited=True):
        """
        Return RoleAssignments for members of project excluding owner and
        delegates.

        :param inherited: Include inherited roles (boolean)
        :return: List of RoleAssignments
        """
        return self.get_roles(
            inherited=inherited,
            min_rank=Role.objects.get(name=PROJECT_ROLE_CONTRIBUTOR).rank,
        )

    def has_role(self, user):
        """
        Return whether user has roles in Project. Returns True if user has local
        role, inherits a role from a parent category, or if public guest access
        is enabled for the project.

        :param user: User object
        :return: Boolean
        """
        if self.public_guest_access or self.get_role(user=user):
            return True
        return False

    def has_role_in_children(self, user):
        """
        Return True if user has a role in any of the children in the project.
        Also returns true if public guest access is true for any child.

        :param user: User object
        :return: Boolean
        """
        if self.type != PROJECT_TYPE_CATEGORY:
            return False
        # User with role in self has at least the same role in children
        if self.has_role(user):
            return True
        children = self.get_children(flat=True)
        if (
            any([c.public_guest_access for c in children])
            or RoleAssignment.objects.filter(
                user=user, project__in=children
            ).count()
            > 0
        ):
            return True
        return False

    def get_source_site(self):
        """
        Return source site or None if this is a locally defined project.

        :return: RemoteProject object or None
        """
        if (
            settings.PROJECTROLES_SITE_MODE
            == SODAR_CONSTANTS['SITE_MODE_SOURCE']
        ):
            return None
        RemoteProject = apps.get_model('projectroles', 'RemoteProject')
        try:
            return RemoteProject.objects.get(
                project_uuid=self.sodar_uuid,
                site__mode=SODAR_CONSTANTS['SITE_MODE_SOURCE'],
            ).site
        except RemoteProject.DoesNotExist:
            pass
        return None

    def is_remote(self):
        """
        Return True if current project has been retrieved from a remote SODAR
        site.

        :return: Boolean
        """
        if (
            settings.PROJECTROLES_SITE_MODE
            == SODAR_CONSTANTS['SITE_MODE_TARGET']
            and self.get_source_site()
        ):
            return True
        return False

    def is_revoked(self):
        """
        Return True if remote access has been revoked for the project.

        :return: Boolean
        """
        if self.is_remote():
            remote_project = RemoteProject.objects.filter(
                project=self, site=self.get_source_site()
            ).first()
            if (
                remote_project
                and remote_project.level
                == SODAR_CONSTANTS['REMOTE_LEVEL_REVOKED']
            ):
                return True
        return False

    def set_public(self, public=True):
        """Helper for setting value of public_guest_access"""
        if public != self.public_guest_access:
            self.public_guest_access = public
            self.save()
            self._update_public_children()  # Update for parents

    def set_archive(self, status=True):
        """
        Helper for setting archive value. Raises ValidationError for categories.
        """
        if status != self.archive:
            self.archive = status
            self.save()

    def get_log_title(self, full_title=False):
        """
        Return a logger-friendly title for the project.

        :param full_title: Display full title if True (boolean)
        :return: String
        """
        return '"{}" ({})'.format(
            self.full_title if full_title else self.title, self.sodar_uuid
        )


# Role -------------------------------------------------------------------------


def get_role_project_type_default():
    """Return default value for Role.project_type"""
    return [PROJECT_TYPE_CATEGORY, PROJECT_TYPE_PROJECT]


class Role(models.Model):
    """Role definition, used to assign roles to projects in RoleAssignment"""

    #: Name of role
    name = models.CharField(
        max_length=64, unique=True, help_text='Name of role'
    )

    #: Role rank for determining role hierarchy
    rank = models.IntegerField(
        default=0,  # 0 = No rank
        help_text='Role rank for determining role hierarchy',
    )

    #: Allowed project types for the role
    project_types = ArrayField(
        models.CharField(max_length=64, blank=False),
        default=get_role_project_type_default,
        help_text='Allowed project types for the role',
    )

    #: Role description
    description = models.TextField(help_text='Role description')

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'Role({})'.format(repr(self.name))


# RoleAssignment ---------------------------------------------------------------


class RoleAssignment(models.Model):
    """
    Assignment of an user to a role in a project. One local assignment per user
    is allowed for each project.

    One local project owner assignment is allowed for a project. Local project
    delegate assignements might be limited using PROJECTROLES_DELEGATE_LIMIT.

    Inherited role assignments can be accessed via the Project model API, e.g.
    Project.get_roles().
    """

    #: Project in which role is assigned
    project = models.ForeignKey(
        Project,
        related_name='local_roles',
        help_text='Project in which role is assigned',
        on_delete=models.CASCADE,
    )

    #: User for whom role is assigned
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        related_name='roles',
        help_text='User for whom role is assigned',
        on_delete=models.CASCADE,
    )

    #: Role to be assigned
    role = models.ForeignKey(
        Role,
        related_name='assignments',
        help_text='Role to be assigned',
        on_delete=models.CASCADE,
    )

    #: RoleAssignment SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='RoleAssignment SODAR UUID'
    )

    class Meta:
        ordering = [
            'project__parent__title',
            'project__title',
            'role__name',
            'user__username',
        ]

    def __str__(self):
        return '{}: {}: {}'.format(self.project, self.role, self.user)

    def __repr__(self):
        values = (self.project.title, self.user.username, self.role.name)
        return 'RoleAssignment({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Version of save() to include custom validation for RoleAssignment"""
        self._validate_project_type()
        self._validate_user()
        self._validate_owner()
        self._validate_delegate()
        super().save(*args, **kwargs)

    def _validate_project_type(self):
        """Validate type of project to ensure it is in allowed types"""
        if self.project.type not in self.role.project_types:
            raise ValidationError(
                ROLE_PROJECT_TYPE_ERROR_MSG.format(
                    project_type=self.project.type, role_name=self.role.name
                )
            )

    def _validate_user(self):
        """
        Validate fields to ensure user has only one role set for the project.
        """
        role_as = RoleAssignment.objects.filter(
            project=self.project, user=self.user
        ).first()
        if role_as and (not self.pk or role_as.pk != self.pk):
            raise ValidationError(
                'Role {} already set for {} in {}'.format(
                    role_as.role, role_as.user, role_as.project
                )
            )

    def _validate_owner(self):
        """
        Validate role to ensure no more than one project owner is assigned to a
        project.
        """
        if self.role.name == PROJECT_ROLE_OWNER:
            owner = self.project.get_owner()
            if owner and (not self.pk or owner.pk != self.pk):
                raise ValidationError(
                    'User {} already set as owner of {}'.format(
                        owner, self.project
                    )
                )

    def _validate_delegate(self):
        """
        Validate role to ensure no more than the allowed amount of project
        delegates are assigned to a project.
        """
        if (
            self.role.name != PROJECT_ROLE_DELEGATE
            or self.project.is_remote()  # No validation for remote projects
        ):
            return
        del_limit = getattr(settings, 'PROJECTROLES_DELEGATE_LIMIT', 1)
        delegates = self.project.get_delegates(inherited=False)
        if 0 < del_limit <= len(delegates) and (
            not self.pk
            or self.project.local_roles.filter(
                role__name=PROJECT_ROLE_DELEGATE, pk=self.pk
            )
            is None
        ):
            raise ValidationError(
                'The local delegate limit for this project ({}) has already '
                'been reached.'.format(del_limit)
            )


# AppSetting -------------------------------------------------------------------


class AppSettingManager(models.Manager):
    """Manager for custom table-level AppSetting queries"""

    def get_setting_value(
        self, app_name, setting_name, project=None, user=None
    ):
        """
        Return value of setting_name for app_name in project or for user.

        Note that project and/or user must be set.

        :param app_name: App plugin name (string)
        :param setting_name: Name of setting (string)
        :param project: Project object or pk
        :param user: User object or pk
        :return: Value (string)
        :raise: AppSetting.DoesNotExist if setting is not found
        """
        query_parameters = {
            'name': setting_name,
            'project': project,
            'user': user,
        }
        if not app_name == 'projectroles':
            query_parameters['app_plugin__name'] = app_name
        setting = super().get_queryset().get(**query_parameters)
        return setting.get_value()


class AppSetting(models.Model):
    """
    Project and users settings value.

    The settings are defined in the "app_settings" member in a SODAR project
    app's plugin. The scope of each setting can be either "USER" or "PROJECT",
    defined for each setting in app_settings. Project AND user-specific settings
    or settings which don't belong to either are are currently not supported.
    """

    #: App to which the setting belongs
    app_plugin = models.ForeignKey(
        Plugin,
        null=True,
        unique=False,
        related_name='settings',
        help_text='App to which the setting belongs',
        on_delete=models.CASCADE,
    )

    #: Project to which the setting belongs
    project = models.ForeignKey(
        Project,
        null=True,
        blank=True,
        related_name='settings',
        help_text='Project to which the setting belongs',
        on_delete=models.CASCADE,
    )

    #: Project to which the setting belongs
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        null=True,
        blank=True,
        related_name='user_settings',
        help_text='User to which the setting belongs',
        on_delete=models.CASCADE,
    )

    #: Name of the setting
    name = models.CharField(
        max_length=255, unique=False, help_text='Name of the setting'
    )

    #: Type of the setting
    type = models.CharField(
        max_length=64, unique=False, help_text='Type of the setting'
    )

    #: Value of the setting
    value = models.CharField(
        max_length=APP_SETTING_VAL_MAXLENGTH,
        unique=False,
        null=True,
        blank=True,
        help_text='Value of the setting',
    )

    #: Optional JSON value for the setting
    value_json = models.JSONField(
        null=True, default=dict, help_text='Optional JSON value for the setting'
    )

    #: Setting visibility in forms
    user_modifiable = models.BooleanField(
        default=True, help_text='Setting visibility in forms'
    )

    #: AppSetting SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='AppSetting SODAR UUID'
    )

    # Set manager for custom queries
    objects = AppSettingManager()

    class Meta:
        ordering = ['project__title', 'app_plugin__name', 'name']
        unique_together = ['project', 'user', 'app_plugin', 'name']

    def __str__(self):
        plugin_name = (
            self.app_plugin.name if self.app_plugin else 'projectroles'
        )
        if self.project:
            label = self.project.title
        elif self.user:
            label = self.user.username
        else:
            label = 'SITE'
        return '{}: {} / {}'.format(label, plugin_name, self.name)

    def __repr__(self):
        values = (
            self.project.title if self.project else None,
            self.user.username if self.user else None,
            self.app_plugin.name if self.app_plugin else 'projectroles',
            self.name,
        )
        return 'AppSetting({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Version of save() to convert 'value' data according to 'type'"""
        if self.type == 'BOOLEAN':
            self.value = str(int(self.value))
        elif self.type == 'INTEGER':
            self.value = str(self.value)
        super().save(*args, **kwargs)

    # Custom row-level functions

    def get_value(self):
        """Return value of the setting in the format specified in 'type'"""
        if self.type == 'INTEGER':
            return int(self.value)
        elif self.type == 'BOOLEAN':
            return bool(int(self.value))
        elif self.type == 'JSON':
            return self.value_json
        return self.value


# ProjectInvite ----------------------------------------------------------------


class ProjectInvite(models.Model):
    """
    Invite which is sent to a non-logged in user, determining their role in
    the project.
    """

    #: Email address of the person to be invited
    email = models.EmailField(
        unique=False,
        null=False,
        blank=False,
        help_text='Email address of the person to be invited',
    )

    #: Project to which the person is invited
    project = models.ForeignKey(
        Project,
        null=False,
        related_name='invites',
        help_text='Project to which the person is invited',
        on_delete=models.CASCADE,
    )

    #: Role assigned to the person
    role = models.ForeignKey(
        Role,
        null=False,
        help_text='Role assigned to the person',
        on_delete=models.CASCADE,
    )

    #: User who issued the invite
    issuer = models.ForeignKey(
        AUTH_USER_MODEL,
        null=False,
        related_name='issued_invites',
        help_text='User who issued the invite',
        on_delete=models.CASCADE,
    )

    #: DateTime of invite creation
    date_created = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of invite creation'
    )

    #: Expiration of invite as DateTime
    date_expire = models.DateTimeField(
        null=False, help_text='Expiration of invite as DateTime'
    )

    #: Message to be included in the invite email (optional)
    message = models.TextField(
        blank=True,
        help_text='Message to be included in the invite email (optional)',
    )

    #: Secret token provided to user with the invite
    secret = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=False,
        help_text='Secret token provided to user with the invite',
    )

    #: Status of the invite (False if claimed or revoked)
    active = models.BooleanField(
        default=True,
        help_text='Status of the invite (False if claimed or revoked)',
    )

    #: ProjectInvite SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='ProjectInvite SODAR UUID'
    )

    class Meta:
        ordering = ['project__title', 'email', 'role__name']

    def __str__(self):
        return '{}: {} ({}){}'.format(
            self.project,
            self.email,
            self.role.name,
            ' [ACTIVE]' if self.active else '',
        )

    def __repr__(self):
        values = (self.project.title, self.email, self.role.name, self.active)
        return 'ProjectInvite({})'.format(', '.join(repr(v) for v in values))


# RemoteSite -------------------------------------------------------------------


class RemoteSite(models.Model):
    """Remote SODAR site"""

    #: Site name
    name = models.CharField(
        max_length=255,
        unique=True,
        blank=False,
        null=False,
        help_text='Site name',
    )

    #: Site URL
    url = models.URLField(
        max_length=2000,
        blank=False,
        null=False,
        unique=False,
        help_text='Site URL',
    )

    #: Site mode
    mode = models.CharField(
        max_length=64,
        unique=False,
        blank=False,
        null=False,
        default=SODAR_CONSTANTS['SITE_MODE_TARGET'],
        help_text='Site mode',
    )

    #: Site description
    description = models.TextField(help_text='Site description')

    #: Secret token used to connect to the master site
    secret = models.CharField(
        max_length=255,
        unique=False,
        blank=False,
        null=True,  # Can be NULL for Peer Mode
        help_text='Secret token for connecting to the source site',
    )

    #: RemoteSite relation UUID (local)
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4,
        unique=True,
        help_text='RemoteSite relation UUID (local)',
    )

    #: RemoteSite's link visibilty for users
    user_display = models.BooleanField(
        default=True, unique=False, help_text='RemoteSite visibility to users'
    )

    class Meta:
        ordering = ['name']
        unique_together = ['url', 'mode', 'secret']

    def __str__(self):
        return '{} ({})'.format(self.name, self.mode)

    def __repr__(self):
        values = (self.name, self.mode, self.url)
        return 'RemoteSite({})'.format(', '.join(repr(v) for v in values))

    def save(self, *args, **kwargs):
        """Version of save() to include custom validation"""
        self._validate_mode()
        super().save(*args, **kwargs)

    def _validate_mode(self):
        """Validate mode value"""
        if self.mode not in SODAR_CONSTANTS['SITE_MODES']:
            raise ValidationError(
                'Mode "{}" not found in SITE_MODES'.format(self.mode)
            )

    # Custom row-level functions

    def get_access_date(self):
        """Return date of latest project access by remote site"""
        projects = RemoteProject.objects.filter(site=self).order_by(
            '-date_access'
        )
        if projects.count() > 0:
            return projects.first().date_access

    def get_url(self):
        """Return sanitized site URL"""
        if self.url[-1] == '/':
            return self.url[:-1]
        return self.url


# RemoteProject ----------------------------------------------------------------


class RemoteProject(models.Model):
    """Remote project relation"""

    #: Related project UUID
    project_uuid = models.UUIDField(
        default=None, unique=False, help_text='Project UUID'
    )

    #: Related project object (if created locally)
    project = models.ForeignKey(
        Project,
        related_name='remotes',
        blank=True,
        null=True,
        help_text='Related project object (if created locally)',
        on_delete=models.CASCADE,
    )

    #: Related remote SODAR site
    site = models.ForeignKey(
        RemoteSite,
        null=False,
        related_name='projects',
        help_text='Remote SODAR site',
        on_delete=models.CASCADE,
    )

    #: Project access level
    level = models.CharField(
        max_length=255,
        unique=False,
        blank=False,
        null=False,
        default=SODAR_CONSTANTS['REMOTE_LEVEL_NONE'],
        help_text='Project access level',
    )

    #: DateTime of last access from/to remote site
    date_access = models.DateTimeField(
        null=True,
        auto_now_add=False,
        help_text='DateTime of last access from/to remote site',
    )

    #: RemoteProject relation UUID (local)
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4,
        unique=True,
        help_text='RemoteProject relation UUID (local)',
    )

    class Meta:
        ordering = ['site__name', 'project_uuid']

    def __str__(self):
        return '{}: {} ({})'.format(
            self.site.name, str(self.project_uuid), self.site.mode
        )

    def __repr__(self):
        values = (self.site.name, str(self.project_uuid), self.site.mode)
        return 'RemoteProject({})'.format(', '.join(repr(v) for v in values))

    # Custom row-level functions

    def get_project(self):
        """Get the related Project object"""
        return (
            self.project
            or Project.objects.filter(sodar_uuid=self.project_uuid).first()
        )


# Abstract User Model ----------------------------------------------------------


class SODARUser(AbstractUser):
    """
    SODAR compatible abstract user model. Use this on your SODAR Core based
    site.
    """

    # First Name and Last Name do not cover name patterns
    # around the globe.
    name = models.CharField(_('Name of User'), blank=True, max_length=255)

    #: User SODAR UUID
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='User SODAR UUID'
    )

    class Meta:
        abstract = True
        ordering = ['name', 'username']

    def __str__(self):
        return self.username

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.set_group()

    def get_full_name(self):
        """Return full name or username if not set"""
        if hasattr(self, 'name') and self.name:
            return self.name
        elif self.first_name and self.last_name:
            return '{} {}'.format(self.first_name, self.last_name)
        return self.username

    def get_form_label(self, email=False):
        """
        Return user label with full name, username and optional email.

        :param email: Return email if True (boolean, default=False)
        :return: String
        """
        ret = self.get_full_name()
        if ret != self.username:
            ret += ' ({})'.format(self.username)
        if email and self.email:
            ret += ' <{}>'.format(self.email)
        return ret

    def set_group(self):
        """Set user group based on user name."""
        if self.username.find('@') != -1:
            group_name = self.username.split('@')[1].lower()
        else:
            group_name = SODAR_CONSTANTS['SYSTEM_USER_GROUP']
        group, created = Group.objects.get_or_create(name=group_name)
        if group not in self.groups.all():
            group.user_set.add(self)
            return group_name

    def is_local(self):
        return not bool(re.search('@[A-Za-z0-9._-]+$', self.username))

    def update_full_name(self):
        """
        Update full name of user.

        :return: String
        """
        # Save user name from first_name and last_name into name
        if self.name in ['', None] and self.first_name != '':
            self.name = self.first_name + (
                ' ' + self.last_name if self.last_name != '' else ''
            )
        self.save()
        return self.name

    def update_ldap_username(self):
        """
        Update username for an LDAP user.

        :return: String
        """
        # Make domain in username uppercase
        if (
            self.username.find('@') != -1
            and self.username.split('@')[1].islower()
        ):
            u_split = self.username.split('@')
            self.username = u_split[0] + '@' + u_split[1].upper()
            self.save()
        return self.username



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-20 15:29
from __future__ import unicode_literals

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0004_rename_uuid'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='sodar_uuid',
            field=models.UUIDField(default=uuid.uuid4, help_text='Project SODAR UUID', unique=True),
        ),
        migrations.AlterField(
            model_name='projectinvite',
            name='sodar_uuid',
            field=models.UUIDField(default=uuid.uuid4, help_text='ProjectInvite SODAR UUID', unique=True),
        ),
        migrations.AlterField(
            model_name='projectsetting',
            name='sodar_uuid',
            field=models.UUIDField(default=uuid.uuid4, help_text='ProjectSetting SODAR UUID', unique=True),
        ),
        migrations.AlterField(
            model_name='projectusertag',
            name='sodar_uuid',
            field=models.UUIDField(default=uuid.uuid4, help_text='ProjectUserTag SODAR UUID', unique=True),
        ),
        migrations.AlterField(
            model_name='roleassignment',
            name='sodar_uuid',
            field=models.UUIDField(default=uuid.uuid4, help_text='RoleAssignment SODAR UUID', unique=True),
        ),
    ]



"""UI views for the filesfolders app"""

import logging

from wsgiref.util import FileWrapper  # For db files
from zipfile import ZipFile

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile
from django.db import transaction
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import (
    TemplateView,
    UpdateView,
    CreateView,
    DeleteView,
    View,
)
from django.views.generic.edit import ModelFormMixin, DeletionMixin

from db_file_storage.storage import DatabaseFileStorage

# Projectroles dependency
from projectroles.models import Project, SODAR_CONSTANTS
from projectroles.plugins import get_backend_api
from projectroles.app_settings import AppSettingAPI
from projectroles.utils import build_secret, get_display_name
from projectroles.views import (
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectContextMixin,
    HTTPRefererMixin,
    ProjectPermissionMixin,
    CurrentUserFormMixin,
    InvalidFormMixin,
)

from filesfolders.forms import FolderForm, FileForm, HyperLinkForm
from filesfolders.models import Folder, File, FileData, HyperLink
from filesfolders.utils import build_public_url


app_settings = AppSettingAPI()
logger = logging.getLogger(__name__)
storage = DatabaseFileStorage()


# Local constants
APP_NAME = 'filesfolders'
TL_OBJ_TYPES = {'Folder': 'folder', 'File': 'file', 'HyperLink': 'hyperlink'}
DEFAULT_UPDATE_ATTRS = ['name', 'folder', 'description', 'flag']
LINK_BAD_REQUEST_MSG = settings.FILESFOLDERS_LINK_BAD_REQUEST_MSG
SERVE_AS_ATTACHMENT = settings.FILESFOLDERS_SERVE_AS_ATTACHMENT


# Mixins -----------------------------------------------------------------


class ObjectPermissionMixin(LoggedInPermissionMixin):
    """Mixin to ensure owner permission for different filesfolders objects"""

    def has_permission(self):
        """Override has_permission to check perms depending on owner"""
        try:
            obj = type(self.get_object()).objects.get(
                sodar_uuid=self.kwargs['item']
            )
            if obj.owner == self.request.user:
                return self.request.user.has_perm(
                    'filesfolders.update_data_own', self.get_permission_object()
                )
            return self.request.user.has_perm(
                'filesfolders.update_data_all', self.get_permission_object()
            )
        except type(self.get_object()).DoesNotExist:
            return False

    def get_permission_object(self):
        """Override get_permission_object for checking Project permission"""
        if self.get_object():
            return self.get_object().project
        return None


class FilesfoldersTimelineMixin:
    """Mixin for filesfolders specific timeline helpers"""

    @classmethod
    def add_item_modify_event(
        cls,
        obj,
        request,
        view_action,
        update_attrs=DEFAULT_UPDATE_ATTRS,
        old_data=None,
    ):
        """
        Add filesfolders item create/update event to timeline.

        :param obj: Filesfolders object being created or updated
        :param request: Request object
        :param view_action: "create" or "update" (string)
        :param update_attrs: List of attribute names to include in extra_data
        :param old_data: Data from existing object in case of update (dict)
        """
        timeline = get_backend_api('timeline_backend')
        if not timeline:
            return

        obj_type = TL_OBJ_TYPES[obj.__class__.__name__]
        extra_data = {}
        tl_desc = '{} {} {{{}}}'.format(view_action, obj_type, obj_type)

        if view_action == 'create':
            for a in update_attrs:
                extra_data[a] = str(getattr(obj, a))

        elif old_data:  # Update
            for a in update_attrs:
                if old_data[a] != getattr(obj, a):
                    extra_data[a] = str(getattr(obj, a))
            tl_desc += ' (' + ', '.join(a for a in extra_data) + ')'

        tl_event = timeline.add_event(
            project=obj.project,
            app_name=APP_NAME,
            user=request.user,
            event_name='{}_{}'.format(obj_type, view_action),
            description=tl_desc,
            extra_data=extra_data,
            status_type='OK',
        )
        tl_event.add_object(
            obj=obj,
            label=obj_type,
            name=obj.get_path() if isinstance(obj, Folder) else obj.name,
        )


class ViewActionMixin(object):
    """Mixin for retrieving form action type"""

    @property
    def view_action(self):
        raise ImproperlyConfigured('Property "view_action" missing!')

    def get_view_action(self):
        return self.view_action if self.view_action else None


class FormValidMixin(ModelFormMixin, FilesfoldersTimelineMixin):
    """Mixin for overriding form_valid in form views for creation/updating"""

    def form_valid(self, form):
        view_action = self.get_view_action()
        old_data = {}
        update_attrs = ['name', 'folder', 'description', 'flag']
        if view_action == 'update':
            old_item = self.get_object()
            if old_item.__class__.__name__ == 'HyperLink':
                update_attrs.append('url')
            elif old_item.__class__.__name__ == 'File':
                update_attrs.append('public_url')
            # Get old fields
            for a in update_attrs:
                old_data[a] = getattr(old_item, a)
        self.object = form.save()

        # Add event in Timeline
        self.add_item_modify_event(
            obj=self.object,
            request=self.request,
            view_action=view_action,
            update_attrs=update_attrs,
            old_data=old_data,
        )
        messages.success(
            self.request,
            '{} "{}" successfully {}d.'.format(
                self.object.__class__.__name__, self.object.name, view_action
            ),
        )
        # TODO: Repetition, put this in a mixin?
        if self.object.folder:
            re_kwargs = {'folder': self.object.folder.sodar_uuid}
        else:
            re_kwargs = {'project': self.object.project.sodar_uuid}
        return redirect(reverse('filesfolders:list', kwargs=re_kwargs))


class DeleteSuccessMixin(DeletionMixin):
    """Mixin for overriding get_success_url in deletion form views"""

    def get_success_url(self):
        timeline = get_backend_api('timeline_backend')
        # Add event in Timeline
        if timeline:
            obj_type = TL_OBJ_TYPES[self.object.__class__.__name__]
            # Add event in Timeline
            tl_event = timeline.add_event(
                project=self.object.project,
                app_name=APP_NAME,
                user=self.request.user,
                event_name='{}_delete'.format(obj_type),
                description='delete {} {{{}}}'.format(obj_type, obj_type),
                status_type='OK',
            )
            tl_event.add_object(
                obj=self.object,
                label=obj_type,
                name=self.object.get_path()
                if isinstance(self.object, Folder)
                else self.object.name,
            )

        messages.success(
            self.request,
            '{} "{}" deleted.'.format(
                self.object.__class__.__name__, self.object.name
            ),
        )
        # TODO: Repetition, put this in a mixin?
        if self.object.folder:
            re_kwargs = {'folder': self.object.folder.sodar_uuid}
        else:
            re_kwargs = {'project': self.object.project.sodar_uuid}
        return reverse('filesfolders:list', kwargs=re_kwargs)


class FileServeMixin:
    """Mixin for file download serving"""

    def get(self, *args, **kwargs):
        """GET request to return the file as attachment"""
        timeline = get_backend_api('timeline_backend')
        if kwargs.get('project'):
            redirect_url = reverse(
                'filesfolders:list', kwargs={'project': kwargs['project']}
            )
        else:
            redirect_url = reverse('home')

        # Get File object
        try:
            file = File.objects.get(sodar_uuid=kwargs['file'])
        except File.DoesNotExist:
            messages.error(self.request, 'File object not found.')
            return redirect(redirect_url)
        # Get corresponding FileData object with file content
        try:
            file_data = FileData.objects.get(file_name=file.file.name)
        except FileData.DoesNotExist:
            messages.error(self.request, 'File data not found.')
            return redirect(redirect_url)
        # Open file for serving
        try:
            file_content = storage.open(file_data.file_name)
        except Exception:
            messages.error(self.request, 'Error opening file.')
            return redirect(redirect_url)

        # Return file as attachment
        response = HttpResponse(
            FileWrapper(file_content), content_type=file_data.content_type
        )
        if SERVE_AS_ATTACHMENT:
            response['Content-Disposition'] = 'attachment; filename={}'.format(
                file.name
            )
        if self.request.user.is_authenticated:
            # Add event in Timeline
            if timeline:
                tl_event = timeline.add_event(
                    project=file.project,
                    app_name=APP_NAME,
                    user=self.request.user,
                    event_name='file_serve',
                    description='serve file {file}',
                    classified=True,
                    status_type='INFO',
                )
                tl_event.add_object(file, 'file', file.name)
        return response


# Base Views -------------------------------------------------------------


class BaseCreateView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    FormValidMixin,
    ProjectContextMixin,
    ProjectPermissionMixin,
    CurrentUserFormMixin,
    InvalidFormMixin,
    CreateView,
):
    """Base File/Folder/HyperLink creation view"""

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        if 'folder' in self.kwargs:
            context['folder'] = Folder.objects.filter(
                sodar_uuid=self.kwargs['folder']
            ).first()
        return context

    def get_form_kwargs(self):
        """Pass current user and URL kwargs to form"""
        kwargs = super().get_form_kwargs()
        if 'folder' in self.kwargs:
            kwargs.update({'folder': self.kwargs['folder']})
        elif 'project' in self.kwargs:
            kwargs.update({'project': self.kwargs['project']})
        return kwargs


# File List View ---------------------------------------------------------


class ProjectFileView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectPermissionMixin,
    ProjectContextMixin,
    TemplateView,
):
    """View for displaying files and folders for a project"""

    # Projectroles dependency
    permission_required = 'filesfolders.view_data'
    template_name = 'filesfolders/project_files.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        project = self.get_project(self.request, self.kwargs)
        context['project'] = project
        # Get folder and file data
        root_folder = None

        if 'folder' in self.kwargs:
            root_folder = Folder.objects.filter(
                sodar_uuid=self.kwargs['folder']
            ).first()
            if root_folder:
                context['folder'] = root_folder
                # Build breadcrumb
                f = root_folder
                breadcrumb = [f]
                while f.folder:
                    breadcrumb.insert(0, f.folder)
                    f = f.folder
                context['folder_breadcrumb'] = breadcrumb
        context['folders'] = Folder.objects.filter(
            project=project, folder=root_folder
        )
        context['files'] = File.objects.filter(
            project=project, folder=root_folder
        )
        context['links'] = HyperLink.objects.filter(
            project=project, folder=root_folder
        )
        folder_pk = (
            Folder.objects.get(sodar_uuid=self.kwargs['folder']).pk
            if 'folder' in self.kwargs
            else None
        )

        # Get folder ReadMe
        readme_md = File.objects.get_folder_readme(
            project_pk=project.pk, folder_pk=folder_pk, mimetype='text/markdown'
        )
        # If the markdown version is not found, try to get a plaintext version
        readme_txt = File.objects.get_folder_readme(
            project_pk=project.pk, folder_pk=folder_pk, mimetype='text/plain'
        )
        readme_file = readme_md if readme_md else readme_txt
        if readme_file:
            try:
                context['readme_name'] = readme_file.name
                context['readme_mime'] = readme_file.file.file.mimetype
                if context['readme_mime'] == 'text/markdown':
                    context['readme_data'] = readme_file.file.read().decode(
                        'utf-8'
                    )
                    if readme_txt:
                        context['readme_alt'] = readme_txt.name
                else:
                    context['readme_data'] = readme_file.file.read()
            except Exception as ex:
                if settings.DEBUG:
                    raise ex
                logger.error(
                    'Exception in accessing readme file data (UUID={}): '
                    '{}'.format(readme_file.sodar_uuid, ex)
                )
        return context


# Folder Views -----------------------------------------------------------


class FolderCreateView(ViewActionMixin, BaseCreateView):
    """Folder creation view"""

    permission_required = 'filesfolders.add_data'
    model = Folder
    form_class = FolderForm
    view_action = 'create'


class FolderUpdateView(
    LoginRequiredMixin,
    ObjectPermissionMixin,
    FormValidMixin,
    ViewActionMixin,
    ProjectContextMixin,
    InvalidFormMixin,
    UpdateView,
):
    """Folder updating view"""

    model = Folder
    form_class = FolderForm
    view_action = 'update'
    slug_url_kwarg = 'item'
    slug_field = 'sodar_uuid'


class FolderDeleteView(
    LoginRequiredMixin,
    ObjectPermissionMixin,
    DeleteSuccessMixin,
    ProjectContextMixin,
    DeleteView,
):
    """Folder deletion view"""

    model = Folder
    slug_url_kwarg = 'item'
    slug_field = 'sodar_uuid'


# File Views -------------------------------------------------------------


class FileCreateView(ViewActionMixin, BaseCreateView):
    """File creation view"""

    permission_required = 'filesfolders.add_data'
    model = File
    form_class = FileForm
    view_action = 'create'

    def form_valid(self, form):
        """Override form_valid() for zip file unpacking"""
        timeline = get_backend_api('timeline_backend')

        # Regular file upload
        if not form.cleaned_data.get('unpack_archive'):
            return super().form_valid(form)

        # Zip file unpacking
        file = form.cleaned_data.get('file')
        folder = form.cleaned_data.get('folder')
        project = self.get_project(self.request, self.kwargs)

        # Build redirect URL
        # TODO: Repetition, put this in a mixin?
        if folder:
            re_kwargs = {'folder': folder.sodar_uuid}
        else:
            re_kwargs = {'project': project.sodar_uuid}
        redirect_url = reverse('filesfolders:list', kwargs=re_kwargs)

        try:
            zip_file = ZipFile(file)
        except Exception as ex:
            messages.error(
                self.request, 'Unable to extract zip file: {}'.format(ex)
            )
            return redirect(redirect_url)

        new_folders = []
        new_files = []

        with transaction.atomic():
            for f in [f for f in zip_file.infolist() if not f.is_dir()]:
                # Create subfolders if any
                current_folder = folder
                for zip_folder in f.filename.split('/')[:-1]:
                    try:
                        current_folder = Folder.objects.get(
                            name=zip_folder,
                            project=project,
                            folder=current_folder,
                        )
                    except Folder.DoesNotExist:
                        current_folder = Folder.objects.create(
                            name=zip_folder,
                            project=project,
                            folder=current_folder,
                            owner=self.request.user,
                        )
                        new_folders.append(current_folder)

                # Save file
                file_name_nopath = f.filename.split('/')[-1]
                unpacked_file = File(
                    name=file_name_nopath,
                    project=project,
                    folder=current_folder,
                    owner=self.request.user,
                    secret=build_secret(),
                )
                content_file = ContentFile(zip_file.read(f.filename))
                unpacked_file.file.save(file_name_nopath, content_file)
                unpacked_file.save()
                new_files.append(unpacked_file)

        # Add timeline events
        for new_folder in new_folders:
            self.add_item_modify_event(
                obj=new_folder, request=self.request, view_action='create'
            )
        for new_file in new_files:
            self.add_item_modify_event(
                obj=new_file, request=self.request, view_action='create'
            )
        if timeline:
            timeline.add_event(
                project=project,
                app_name=APP_NAME,
                user=self.request.user,
                event_name='archive_extract',
                description='Extract from archive "{}", create {} folders '
                'and {} files'.format(
                    file.name, len(new_folders), len(new_files)
                ),
                extra_data={
                    'new_folders': [f.name for f in new_folders],
                    'new_files': [f.name for f in new_files],
                },
                status_type='OK',
            )

        messages.success(
            self.request,
            'Extracted {} files in folder "{}" from archive "{}".'.format(
                len([f for f in zip_file.infolist() if not f.is_dir()]),
                folder.name if folder else 'root',
                file.name,
            ),
        )
        return redirect(redirect_url)


class FileUpdateView(
    LoginRequiredMixin,
    ObjectPermissionMixin,
    FormValidMixin,
    ViewActionMixin,
    ProjectContextMixin,
    InvalidFormMixin,
    UpdateView,
):
    """File updating view"""

    model = File
    form_class = FileForm
    view_action = 'update'
    slug_url_kwarg = 'item'
    slug_field = 'sodar_uuid'


class FileDeleteView(
    LoginRequiredMixin,
    ObjectPermissionMixin,
    DeleteSuccessMixin,
    ProjectContextMixin,
    DeleteView,
):
    """File deletion view"""

    model = File
    slug_url_kwarg = 'item'
    slug_field = 'sodar_uuid'


class FileServeView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    FileServeMixin,
    ProjectPermissionMixin,
    View,
):
    """View for serving file to a logged in user with permissions"""

    permission_required = 'filesfolders.view_data'


class FileServePublicView(FileServeMixin, View):
    """View for serving file to a public user with secure link"""

    def get(self, *args, **kwargs):
        """Override of GET for checking request URL"""
        try:
            file = File.objects.get(secret=kwargs['secret'])
            # Check if sharing public files is not allowed in project settings
            if not app_settings.get(
                APP_NAME, 'allow_public_links', file.project
            ):
                return HttpResponseBadRequest(LINK_BAD_REQUEST_MSG)
        except File.DoesNotExist:
            return HttpResponseBadRequest(LINK_BAD_REQUEST_MSG)

        # If public URL serving is disabled, don't serve file
        if not file.public_url:
            return HttpResponseBadRequest(LINK_BAD_REQUEST_MSG)
        # Update kwargs with file and project uuid:s
        kwargs.update(
            {'file': file.sodar_uuid, 'project': file.project.sodar_uuid}
        )
        # If successful, return get() from FileServeMixin
        return super().get(*args, **kwargs)


class FilePublicLinkView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    ProjectContextMixin,
    ProjectPermissionMixin,
    TemplateView,
):
    """View for generating a public secure link to a file"""

    permission_required = 'filesfolders.share_public_link'
    template_name = 'filesfolders/public_link.html'

    def get(self, *args, **kwargs):
        """Override of GET for checking project settings"""
        try:
            file = File.objects.get(sodar_uuid=self.kwargs['file'])
        except File.DoesNotExist:
            messages.error(self.request, 'File not found.')
            return redirect(reverse('home'))
        if not app_settings.get(APP_NAME, 'allow_public_links', file.project):
            messages.error(
                self.request,
                'Sharing public links not allowed for this {}.'.format(
                    get_display_name(SODAR_CONSTANTS['PROJECT_TYPE_PROJECT'])
                ),
            )
            return redirect(
                reverse(
                    'filesfolders:list',
                    kwargs={'project': file.project.sodar_uuid},
                )
            )
        return super().get(*args, **kwargs)

    def get_context_data(self, *args, **kwargs):
        """Provide URL to context"""
        context = super().get_context_data(*args, **kwargs)
        try:
            file = File.objects.get(sodar_uuid=self.kwargs['file'])
        except File.DoesNotExist:
            messages.error(self.request, 'File not found.')
            return redirect(reverse('home'))
        if not file.public_url:
            messages.error(self.request, 'Public URL for file not enabled.')
            return redirect(
                reverse(
                    'filesfolders:list',
                    kwargs={'project': file.project.sodar_uuid},
                )
            )
        context['file'] = file
        context['public_url'] = build_public_url(file, self.request)
        return context


# HyperLink Views --------------------------------------------------------


class HyperLinkCreateView(ViewActionMixin, BaseCreateView):
    """HyperLink creation view"""

    permission_required = 'filesfolders.add_data'
    model = HyperLink
    form_class = HyperLinkForm
    view_action = 'create'


class HyperLinkUpdateView(
    LoginRequiredMixin,
    ObjectPermissionMixin,
    FormValidMixin,
    ViewActionMixin,
    ProjectContextMixin,
    InvalidFormMixin,
    UpdateView,
):
    """HyperLink updating view"""

    model = HyperLink
    form_class = HyperLinkForm
    view_action = 'update'
    slug_url_kwarg = 'item'
    slug_field = 'sodar_uuid'


class HyperLinkDeleteView(
    LoginRequiredMixin,
    ObjectPermissionMixin,
    DeleteSuccessMixin,
    ProjectContextMixin,
    DeleteView,
):
    """HyperLink deletion view"""

    model = HyperLink
    slug_url_kwarg = 'item'
    slug_field = 'sodar_uuid'


# Batch Edit Views --------------------------------------------------------


class BatchEditView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    HTTPRefererMixin,
    ProjectPermissionMixin,
    TemplateView,
):
    """Batch delete/move confirm view"""

    http_method_names = ['post']
    template_name = 'filesfolders/batch_edit_confirm.html'
    # NOTE: minimum perm, all checked files will be tested in post()
    permission_required = 'filesfolders.update_data_own'

    #: Items we will delete
    items = None

    #: Item IDs to be deleted (so we don't have to create them again)
    item_names = None

    #: Items which we can't delete
    failed = None

    #: Current action of view
    batch_action = None

    #: Project object
    project = None

    def _render_confirmation(self, **kwargs):
        """Render user confirmation"""
        context = {
            'batch_action': self.batch_action,
            'items': self.items,
            'item_names': self.item_names,
            'failed': self.failed,
            'project': self.project,
            'folder_check': True,
        }
        if 'folder' in kwargs:
            context['folder'] = kwargs['folder']

        # NOTE: No modifications needed for "delete" action
        if self.batch_action == 'move':
            # Exclude folders to be moved
            exclude_list = [
                x.sodar_uuid for x in self.items if isinstance(x, Folder)
            ]
            # Exclude folders under folders to be moved
            for i in self.items:
                exclude_list += [
                    x.sodar_uuid
                    for x in Folder.objects.filter(
                        project__sodar_uuid=self.project.sodar_uuid
                    )
                    if x.has_in_path(i)
                ]
            # Exclude current folder
            if 'folder' in kwargs:
                exclude_list.append(kwargs['folder'])
            folder_choices = Folder.objects.filter(
                project=self.project
            ).exclude(sodar_uuid__in=exclude_list)
            context['folder_choices'] = folder_choices
            if folder_choices.count() == 0:
                context['folder_check'] = False
        return super().render_to_response(context)

    def _finalize_edit(self, edit_count, target_folder, **kwargs):
        """Finalize executed batch operation"""
        timeline = get_backend_api('timeline_backend')
        edit_suffix = 's' if edit_count != 1 else ''
        fail_suffix = 's' if len(self.failed) != 1 else ''

        if len(self.failed) > 0:
            messages.warning(
                self.request,
                'Unable to edit {} item{}, check '
                'permissions and target folder. Failed: {}'.format(
                    len(self.failed),
                    fail_suffix,
                    ', '.join(f.name for f in self.failed),
                ),
            )
        if edit_count > 0:
            messages.success(
                self.request,
                'Batch {} {} item{}.'.format(
                    'deleted' if self.batch_action == 'delete' else 'moved',
                    edit_count,
                    edit_suffix,
                ),
            )

        # Add event in Timeline
        if timeline:
            extra_data = {
                'items': [x.name for x in self.items],
                'failed': [x.name for x in self.failed],
            }
            tl_event = timeline.add_event(
                project=Project.objects.filter(
                    sodar_uuid=self.project.sodar_uuid
                ).first(),
                app_name=APP_NAME,
                user=self.request.user,
                event_name='batch_{}'.format(self.batch_action),
                description='batch {} {} item{} {} {}'.format(
                    self.batch_action,
                    edit_count,
                    edit_suffix,
                    '({} failed)'.format(len(self.failed))
                    if len(self.failed) > 0
                    else '',
                    'to {target_folder}'
                    if self.batch_action == 'move' and target_folder
                    else '',
                ),
                extra_data=extra_data,
                status_type='OK' if edit_count > 0 else 'FAILED',
            )
            if self.batch_action == 'move' and target_folder:
                tl_event.add_object(
                    target_folder, 'target_folder', target_folder.get_path()
                )

        if 'folder' in kwargs:
            re_kwargs = {'folder': kwargs['folder']}
        else:
            re_kwargs = {'project': kwargs['project']}
        return redirect(reverse('filesfolders:list', kwargs=re_kwargs))

    def post(self, request, **kwargs):
        """Handle POST request for modifying items or user confirmation"""
        post_data = request.POST
        self.project = self.get_project(request, kwargs)
        self.batch_action = post_data['batch-action']
        self.items = []
        self.item_names = []
        self.failed = []
        can_update_all = request.user.has_perm(
            'filesfolders.update_data_all', self.get_permission_object()
        )
        user_confirmed = bool(int(post_data['user-confirmed']))
        edit_count = 0
        target_folder = None

        if (
            self.batch_action == 'move'
            and 'target-folder' in post_data
            and post_data['target-folder'] != '0'
        ):
            target_folder = Folder.objects.filter(
                sodar_uuid=post_data['target-folder']
            ).first()

        for key in [
            key
            for key, val in post_data.items()
            if key.startswith('batch_item') and val == '1'
        ]:
            cls = eval(key.split('_')[2])
            item = cls.objects.filter(sodar_uuid=key.split('_')[3]).first()
            #: Item permission
            perm_ok = can_update_all | (item.owner == request.user)

            # Checks
            # Perm check
            if not perm_ok:
                self.failed.append(item)
            # Moving checks (after user has selected target folder)
            elif self.batch_action == 'move' and user_confirmed:
                # Can't move if item with same name in target
                get_kwargs = {
                    'project': self.project,
                    'folder': target_folder if target_folder else None,
                    'name': item.name,
                }
                if cls.objects.filter(**get_kwargs):
                    self.failed.append(item)
            # Deletion checks
            elif self.batch_action == 'delete':
                # Can't delete a non-empty folder
                if isinstance(item, Folder) and not item.is_empty():
                    self.failed.append(item)

            # Modify item
            if perm_ok and item not in self.failed:
                if not user_confirmed:
                    self.items.append(item)
                    self.item_names.append(key)
                elif self.batch_action == 'move':
                    item.folder = target_folder
                    item.save()
                    edit_count += 1
                elif self.batch_action == 'delete':
                    item.delete()
                    edit_count += 1

        # Render/redirect
        # Confirmation needed
        if not user_confirmed:
            return self._render_confirmation(**kwargs)
        # User confirmed, batch operation done
        else:
            return self._finalize_edit(edit_count, target_folder, **kwargs)



"""Tests for UI views in the adminalerts app"""
from django.urls import reverse
from django.utils import timezone

from test_plus.test import TestCase

from adminalerts.models import AdminAlert
from adminalerts.tests.test_models import AdminAlertMixin


class TestViewsBase(AdminAlertMixin, TestCase):
    """Base class for view testing"""

    def setUp(self):
        # Create users
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        # No user
        self.anonymous = None
        # Create alert
        self.alert = self.make_alert(
            message='alert',
            user=self.superuser,
            description='description',
            active=True,
            require_auth=True,
        )
        self.expiry_str = (
            timezone.now() + timezone.timedelta(days=1)
        ).strftime('%Y-%m-%d')


class TestAdminAlertListView(TestViewsBase):
    """Tests for the alert list view"""

    def test_render(self):
        """Test rendering of the alert list view"""
        with self.login(self.superuser):
            response = self.client.get(reverse('adminalerts:list'))
        self.assertEqual(response.status_code, 200)
        self.assertIsNotNone(response.context['object_list'])
        self.assertEqual(response.context['object_list'][0].pk, self.alert.pk)


class TestAdminAlertDetailView(TestViewsBase):
    """Tests for the alert detail view"""

    def test_render(self):
        """Test rendering of the alert detail view"""
        with self.login(self.superuser):
            response = self.client.get(
                reverse(
                    'adminalerts:detail',
                    kwargs={'adminalert': self.alert.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['object'], self.alert)


class TestAdminAlertCreateView(TestViewsBase):
    """Tests for the alert creation view"""

    def test_render(self):
        """Test rendering of the alert creation view"""
        with self.login(self.superuser):
            response = self.client.get(reverse('adminalerts:create'))
        self.assertEqual(response.status_code, 200)

    def test_create(self):
        """Test creating an admin alert"""
        self.assertEqual(AdminAlert.objects.all().count(), 1)
        post_data = {
            'message': 'new alert',
            'description': 'description',
            'date_expire': self.expiry_str,
            'active': 1,
            'require_auth': 1,
        }
        with self.login(self.superuser):
            response = self.client.post(
                reverse('adminalerts:create'), post_data
            )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse('adminalerts:list'))
        self.assertEqual(AdminAlert.objects.all().count(), 2)

    def test_create_expired(self):
        """Test creating an admin alert with and old date_expiry (should fail)"""
        self.assertEqual(AdminAlert.objects.all().count(), 1)
        expiry_fail = (timezone.now() + timezone.timedelta(days=-1)).strftime(
            '%Y-%m-%d'
        )
        post_data = {
            'message': 'new alert',
            'description': 'description',
            'date_expire': expiry_fail,
            'active': 1,
            'require_auth': 1,
        }
        with self.login(self.superuser):
            response = self.client.post(
                reverse('adminalerts:create'), post_data
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(AdminAlert.objects.all().count(), 1)


class TestAdminAlertUpdateView(TestViewsBase):
    """Tests for the alert update view"""

    def test_render(self):
        """Test rendering of the alert update view"""
        with self.login(self.superuser):
            response = self.client.get(
                reverse(
                    'adminalerts:update',
                    kwargs={'adminalert': self.alert.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)

    def test_update(self):
        """Test updating an admin alert"""
        self.assertEqual(AdminAlert.objects.all().count(), 1)

        post_data = {
            'message': 'updated alert',
            'description': 'updated description',
            'date_expire': self.expiry_str,
            'active': '',
        }
        with self.login(self.superuser):
            response = self.client.post(
                reverse(
                    'adminalerts:update',
                    kwargs={'adminalert': self.alert.sodar_uuid},
                ),
                post_data,
            )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse('adminalerts:list'))
        self.assertEqual(AdminAlert.objects.all().count(), 1)
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.message, 'updated alert')
        self.assertEqual(self.alert.description.raw, 'updated description')
        self.assertEqual(self.alert.active, False)

    def test_update_user(self):
        """Test updating an admin alert with a different user"""
        superuser2 = self.make_user('superuser2')
        superuser2.is_superuser = True
        superuser2.save()

        post_data = {
            'message': 'updated alert',
            'description': 'updated description',
            'date_expire': self.expiry_str,
            'active': '',
        }
        with self.login(superuser2):
            response = self.client.post(
                reverse(
                    'adminalerts:update',
                    kwargs={'adminalert': self.alert.sodar_uuid},
                ),
                post_data,
            )

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse('adminalerts:list'))
        self.alert.refresh_from_db()
        self.assertEqual(self.alert.user, superuser2)


class TestAdminAlertDeleteView(TestViewsBase):
    """Tests for the alert deletion view"""

    def test_render(self):
        """Test rendering of the alert deletion view"""
        with self.login(self.superuser):
            response = self.client.get(
                reverse(
                    'adminalerts:delete',
                    kwargs={'adminalert': self.alert.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 200)

    def test_delete(self):
        """Test deleting an admin alert"""
        self.assertEqual(AdminAlert.objects.all().count(), 1)
        with self.login(self.superuser):
            response = self.client.post(
                reverse(
                    'adminalerts:delete',
                    kwargs={'adminalert': self.alert.sodar_uuid},
                )
            )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse('adminalerts:list'))
        self.assertEqual(AdminAlert.objects.all().count(), 0)



# Generated by Django 3.2.9 on 2021-11-17 09:34

from django.db import migrations, models


def get_public_children(obj):
    for child in obj.children.all():
        if child.public_guest_access:
            return True
        ret = get_public_children(child)
        if ret:
            return True
    return False


def populate_public_children(apps, schema_editor):
    """Populate the new has_public_children field in the Project model"""
    Project = apps.get_model('projectroles', 'Project')
    for project in Project.objects.all():
        project.has_public_children = get_public_children(project)
        project.save()


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0019_project_public_guest_access'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='has_public_children',
            field=models.BooleanField(default=False, help_text='Whether project has children with public access (auto-generated)'),
        ),
        migrations.RunPython(
            populate_public_children,
            reverse_code=migrations.RunPython.noop
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-19 15:15
from __future__ import unicode_literals

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('timeline', '0002_projectevent_user'),
    ]

    operations = [
        migrations.RenameField(
            model_name='projectevent',
            old_name='omics_uuid',
            new_name='sodar_uuid'
        ),
    ]



from django.urls import path

from userprofile import views

app_name = 'userprofile'

urlpatterns = [
    path(
        route='profile',
        view=views.UserDetailView.as_view(),
        name='detail',
    ),
    path(
        route='profile/settings/update',
        view=views.UserSettingsView.as_view(),
        name='settings_update',
    ),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-19 15:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0003_populate_roles'),
    ]

    operations = [
        migrations.RenameField(
            model_name='project',
            old_name='omics_uuid',
            new_name='sodar_uuid'
        ),
        migrations.RenameField(
            model_name='projectinvite',
            old_name='omics_uuid',
            new_name='sodar_uuid'
        ),
        migrations.RenameField(
            model_name='projectsetting',
            old_name='omics_uuid',
            new_name='sodar_uuid'
        ),
        migrations.RenameField(
            model_name='projectusertag',
            old_name='omics_uuid',
            new_name='sodar_uuid'
        ),
        migrations.RenameField(
            model_name='roleassignment',
            old_name='omics_uuid',
            new_name='sodar_uuid'
        ),
    ]



"""Projectroles app settings API"""

import json
import logging

from django.conf import settings

from projectroles.models import AppSetting, APP_SETTING_TYPES, SODAR_CONSTANTS
from projectroles.plugins import get_app_plugin, get_active_plugins


logger = logging.getLogger(__name__)


# SODAR constants
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']

# Local constants
APP_SETTING_LOCAL_DEFAULT = True
APP_SETTING_SCOPES = [
    APP_SETTING_SCOPE_PROJECT,
    APP_SETTING_SCOPE_USER,
    APP_SETTING_SCOPE_PROJECT_USER,
    APP_SETTING_SCOPE_SITE,
]
APP_SETTING_DEFAULT_VALUES = {
    'STRING': '',
    'INTEGER': 0,
    'BOOLEAN': False,
    'JSON': {},
}
APP_SETTING_SCOPE_ARGS = {
    APP_SETTING_SCOPE_PROJECT: {'project': True, 'user': False},
    APP_SETTING_SCOPE_USER: {'project': False, 'user': True},
    APP_SETTING_SCOPE_PROJECT_USER: {'project': True, 'user': True},
    APP_SETTING_SCOPE_SITE: {'project': False, 'user': False},
}
DELETE_SCOPE_ERR_MSG = 'Argument "{arg}" must be set for {scope} scope setting'
OVERRIDE_ERR_MSG = 'Overriding global settings for remote projects not allowed'


# Define App Settings for projectroles app
PROJECTROLES_APP_SETTINGS = {
    #: App settings definition
    #:
    #: Example ::
    #:
    #:     'example_setting': {
    #:         'scope': 'PROJECT',  # PROJECT/USER/PROJECT-USER/SITE
    #:         'type': 'STRING',  # STRING/INTEGER/BOOLEAN/JSON
    #:         'default': 'example',
    #:         'label': 'Project setting',  # Optional, defaults to name/key
    #:         'placeholder': 'Enter example setting here',  # Optional
    #:         'description': 'Example project setting',  # Optional
    #:         'options': ['example', 'example2'],  # Optional, only for
    #:                    settings of type STRING or INTEGER
    #:         'user_modifiable': True,  # Optional, show/hide in forms
    #:         'local': False,  # Allow editing in target site forms if True
    #:         'project_types': [PROJECT_TYPE_PROJECT],  # Optional, list may
    #:          contain PROJECT_TYPE_CATEGORY and/or PROJECT_TYPE_PROJECT
    #:     }
    'ip_restrict': {
        'scope': APP_SETTING_SCOPE_PROJECT,
        'type': 'BOOLEAN',
        'default': False,
        'label': 'IP restrict',
        'description': 'Restrict project access by an allowed IP list',
        'user_modifiable': True,
        'local': False,
    },
    'ip_allowlist': {
        'scope': APP_SETTING_SCOPE_PROJECT,
        'type': 'JSON',
        'default': [],
        'label': 'IP allow list',
        'description': 'List of allowed IPs for project access',
        'user_modifiable': True,
        'local': False,
    },
    'user_email_additional': {
        'scope': APP_SETTING_SCOPE_USER,
        'type': 'STRING',
        'default': '',
        'placeholder': 'email1@example.com;email2@example.com',
        'label': 'Additional email',
        'description': 'Also send user emails to these addresses. Separate '
        'multiple emails with semicolon.',
        'user_modifiable': True,
        'local': False,
        'project_types': [PROJECT_TYPE_PROJECT],
    },
    'project_star': {
        'scope': APP_SETTING_SCOPE_PROJECT_USER,
        'type': 'BOOLEAN',
        'default': False,
        'local': True,
        'project_types': [PROJECT_TYPE_PROJECT, PROJECT_TYPE_CATEGORY],
    },
}


class AppSettingAPI:
    @classmethod
    def _check_project_and_user(cls, scope, project, user):
        """
        Ensure project and user parameters are set according to scope.

        :param scope: Scope of Setting (USER, PROJECT, PROJECT_USER, SITE)
        :param project: Project object
        :param user: User object
        :raise: ValueError if none or both objects exist
        """
        if not APP_SETTING_SCOPE_ARGS[scope] == {
            'project': project is not None,
            'user': user is not None,
        }:
            raise ValueError(
                'Project and/or user are set incorrect for setting '
                'with {} scope'.format(scope)
            )

    @classmethod
    def _check_scope(cls, scope):
        """
        Ensure the validity of a scope definition.

        :param scope: String
        :raise: ValueError if scope is not recognized
        """
        if scope not in APP_SETTING_SCOPES:
            raise ValueError('Invalid scope "{}"'.format(scope))

    @classmethod
    def _check_type(cls, setting_type):
        """
        Ensure the validity of app setting type.

        :param setting_type: String
        :raise: ValueError if type is not recognized
        """
        if setting_type not in APP_SETTING_TYPES:
            raise ValueError('Invalid setting type "{}"'.format(setting_type))

    @classmethod
    def _check_type_options(cls, setting_type, setting_options):
        """
        Ensure setting_type is allowed to have options.

        :param setting_type: String
        :param setting_options: List of options (Strings or Integers)
        :raise: ValueError if type is not recognized
        """
        if setting_type not in ['INTEGER', 'STRING'] and setting_options:
            raise ValueError(
                'Options are only allowed for settings of type INTEGER and '
                'STRING'
            )

    @classmethod
    def _check_value_in_options(
        cls, setting_value, setting_options, project=None, user=None
    ):
        """
        Ensure setting_value is present in setting_options.

        :param setting_value: String
        :param setting_options: List of options (String or Integers)
        :param project: Project object
        :param user: User object
        :raise: ValueError if type is not recognized
        """
        if callable(setting_options):
            valid_options = [
                val[0] if isinstance(val, tuple) else val
                for val in setting_options(project, user)
            ]
            if setting_value not in valid_options:
                raise ValueError(
                    'Choice "{}" not found in options ({})'.format(
                        setting_value,
                        ', '.join(map(str, valid_options)),
                    )
                )
        elif (
            setting_options
            and not callable(setting_options)
            and setting_value not in setting_options
        ):
            raise ValueError(
                'Choice "{}" not found in options ({})'.format(
                    setting_value, ', '.join(map(str, setting_options))
                )
            )

    @classmethod
    def _get_app_plugin(cls, app_name):
        """
        Return app plugin by name.

        :param app_name: Name of the app plugin (string or None)
        :return: App plugin object
        :raise: ValueError if plugin is not found with the name
        """
        plugin = get_app_plugin(app_name)
        if not plugin:
            raise ValueError(
                'Plugin not found with app name "{}"'.format(app_name)
            )
        return plugin

    @classmethod
    def _get_defs(cls, plugin=None, app_name=None):
        """
        Ensure valid argument values for a settings def query.

        :param plugin: Plugin object or None
        :param app_name: Name of the app plugin (string or None)
        :return: Dict
        :raise: ValueError if args are not valid or plugin is not found
        """
        if not plugin and not app_name:
            raise ValueError('Plugin and app name both unset')
        if app_name == 'projectroles':
            return cls.get_projectroles_defs()
        if not plugin:
            plugin = cls._get_app_plugin(app_name)
        return plugin.app_settings

    @classmethod
    def _get_json_value(cls, value):
        """
        Return JSON value as dict regardless of input type

        :param value: Original value (string or dict)
        :raise: json.decoder.JSONDecodeError if string value is not valid JSON
        :raise: ValueError if value type is not recognized or if value is not
                valid JSON
        :return: dict
        """
        if not value:
            return {}
        try:
            if isinstance(value, str):
                return json.loads(value)
            else:
                json.dumps(value)  # Ensure this is valid
                return value
        except Exception:
            raise ValueError('Value is not valid JSON: {}'.format(value))

    @classmethod
    def _compare_value(cls, obj, input_value):
        """
        Compare input value to value in an AppSetting object. Return True if
        values match, False if there is a mismatch.

        :param obj: AppSetting object
        :param input_value: Input value (string, int, bool or dict)
        :return: Bool
        """
        if obj.type == 'JSON':
            return (
                not obj.value_json and not input_value
            ) or obj.value_json == cls._get_json_value(input_value)
        elif obj.type == 'BOOLEAN':
            return bool(int(obj.value)) == input_value
        return obj.value == str(input_value)

    @classmethod
    def _log_set_debug(
        cls, action, app_name, setting_name, value, project, user
    ):
        """
        Helper method for logging setting changes in set() method.

        :param action: Action string (string)
        :param app_name: Plugin app name (string)
        :param setting_name: Setting name (string)
        :param value: Setting value (string)
        :param project: Project object
        :param user: User object
        """
        extra_data = []
        if project:
            extra_data.append('project={}'.format(project.sodar_uuid))
        if user:
            extra_data.append('user={}'.format(user.username))
        logger.debug(
            '{} app setting: {}.{} = "{}"{}'.format(
                action,
                app_name,
                setting_name,
                value,
                ' ({})'.format('; '.join(extra_data)) if extra_data else '',
            )
        )

    @classmethod
    def get_default(
        cls, app_name, setting_name, project=None, user=None, post_safe=False
    ):
        """
        Get default setting value from an app plugin.

        :param app_name: App name (string, must equal "name" in app plugin)
        :param setting_name: Setting name (string)
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether a POST safe value should be returned (bool)
        :return: Setting value (string, integer or boolean)
        :raise: ValueError if app plugin is not found
        :raise: KeyError if nothing is found with setting_name
        """
        if app_name == 'projectroles':
            app_settings = cls.get_projectroles_defs()
        else:
            app_plugin = get_app_plugin(app_name)
            if not app_plugin:
                raise ValueError('App plugin not found: "{}"'.format(app_name))
            app_settings = app_plugin.app_settings

        if setting_name in app_settings:
            if callable(app_settings[setting_name].get('default')):
                try:
                    callable_setting = app_settings[setting_name]['default']
                    return callable_setting(project, user)
                except Exception:
                    logger.error(
                        'Error in callable setting "{}" for app "{}"'.format(
                            setting_name, app_name
                        )
                    )
                    return APP_SETTING_DEFAULT_VALUES[
                        app_settings[setting_name]['type']
                    ]
            elif app_settings[setting_name]['type'] == 'JSON':
                json_default = app_settings[setting_name].get('default')
                if not json_default:
                    if isinstance(json_default, dict):
                        return {}
                    elif isinstance(json_default, list):
                        return []
                    return {}
                if post_safe:
                    return json.dumps(app_settings[setting_name]['default'])
            return app_settings[setting_name]['default']

        raise KeyError(
            'Setting "{}" not found in app plugin "{}"'.format(
                setting_name, app_name
            )
        )

    @classmethod
    def get(
        cls, app_name, setting_name, project=None, user=None, post_safe=False
    ):
        """
        Return app setting value for a project or a user. If not set, return
        default.

        :param app_name: App name (string, must equal "name" in app plugin)
        :param setting_name: Setting name (string)
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether a POST safe value should be returned (bool)
        :return: String or None
        :raise: KeyError if nothing is found with setting_name
        """
        if not user or user.is_authenticated:
            try:
                val = AppSetting.objects.get_setting_value(
                    app_name, setting_name, project=project, user=user
                )
            except AppSetting.DoesNotExist:
                val = cls.get_default(
                    app_name,
                    setting_name,
                    project=project,
                    user=user,
                    post_safe=post_safe,
                )
        else:  # Anonymous user
            val = cls.get_default(
                app_name,
                setting_name,
                project=project,
                user=user,
                post_safe=post_safe,
            )
        # Handle post_safe for dict values (JSON)
        if post_safe and isinstance(val, (dict, list)):
            return json.dumps(val)
        return val

    @classmethod
    def get_all(cls, project=None, user=None, post_safe=False):
        """
        Return all setting values. If the value is not found, return
        the default.

        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether POST safe values should be returned (bool)
        :return: Dict
        :raise: ValueError if neither project nor user are set
        """
        if not project and not user:
            raise ValueError('Project and user are both unset')

        ret = {}
        app_plugins = get_active_plugins()
        for plugin in app_plugins:
            p_settings = cls.get_definitions(
                APP_SETTING_SCOPE_PROJECT, plugin=plugin
            )
            for s_key in p_settings:
                ret['settings.{}.{}'.format(plugin.name, s_key)] = cls.get(
                    plugin.name, s_key, project, user, post_safe
                )

        p_settings = cls.get_definitions(
            APP_SETTING_SCOPE_PROJECT, app_name='projectroles'
        )
        for s_key in p_settings:
            ret['settings.{}.{}'.format('projectroles', s_key)] = cls.get(
                'projectroles', s_key, post_safe
            )
        return ret

    @classmethod
    def get_defaults(cls, scope, project=None, user=None, post_safe=False):
        """
        Get all default settings for a scope.

        :param scope: Setting scope (PROJECT, USER or PROJECT_USER)
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether POST safe values should be returned (bool)
        :return: Dict
        """
        cls._check_scope(scope)
        ret = {}
        app_plugins = get_active_plugins()

        for plugin in app_plugins:
            p_settings = cls.get_definitions(scope, plugin=plugin)
            for s_key in p_settings:
                ret[
                    'settings.{}.{}'.format(plugin.name, s_key)
                ] = cls.get_default(
                    plugin.name,
                    s_key,
                    project=project,
                    user=user,
                    post_safe=post_safe,
                )

        p_settings = cls.get_definitions(scope, app_name='projectroles')
        for s_key in p_settings:
            ret[
                'settings.{}.{}'.format('projectroles', s_key)
            ] = cls.get_default(
                'projectroles',
                s_key,
                project=project,
                user=user,
                post_safe=post_safe,
            )
        return ret

    @classmethod
    def set(
        cls,
        app_name,
        setting_name,
        value,
        project=None,
        user=None,
        validate=True,
    ):
        """
        Set value of an existing project or user settings. Creates the object if
        not found.

        :param app_name: App name (string, must equal "name" in app plugin)
        :param setting_name: Setting name (string)
        :param value: Value to be set
        :param project: Project object (optional)
        :param user: User object (optional)
        :param validate: Validate value (bool, default=True)
        :return: True if changed, False if not changed
        :raise: ValueError if validating and value is not accepted for setting
                type
        :raise: ValueError if neither project nor user are set
        :raise: KeyError if setting name is not found in plugin specification
        """
        s_def = cls.get_definition(name=setting_name, app_name=app_name)
        cls._check_scope(s_def.get('scope', None))
        cls._check_project_and_user(s_def.get('scope', None), project, user)
        # Check project type
        if (
            project
            and not s_def.get('project_types', None)
            and not project.type == PROJECT_TYPE_PROJECT
            or project
            and s_def.get('project_types', None)
            and project.type not in s_def['project_types']
        ):
            raise ValueError(
                'Project type {} not allowed for setting {}'.format(
                    project.type, setting_name
                )
            )
        # Prevent updating global setting on remote project
        # TODO: Prevent editing global USER settings (#1329)
        if (
            not s_def.get('local', APP_SETTING_LOCAL_DEFAULT)
            and project
            and project.is_remote()
        ):
            raise ValueError(OVERRIDE_ERR_MSG)

        try:  # Update existing setting
            q_kwargs = {'name': setting_name, 'project': project, 'user': user}
            if not app_name == 'projectroles':
                q_kwargs['app_plugin__name'] = app_name
            setting = AppSetting.objects.get(**q_kwargs)
            if cls._compare_value(setting, value):
                return False
            if validate:
                cls.validate(
                    setting.type,
                    value,
                    s_def.get('options'),
                    project=project,
                    user=user,
                )
            if setting.type == 'JSON':
                setting.value_json = cls._get_json_value(value)
            else:
                setting.value = value
            setting.save()
            cls._log_set_debug(
                'Set', app_name, setting_name, value, project, user
            )
            return True

        except AppSetting.DoesNotExist:  # Create new
            s_type = s_def['type']
            if app_name == 'projectroles':
                app_plugin_model = None
            else:
                app_plugin = get_app_plugin(app_name)
                app_plugin_model = app_plugin.get_model()
            if validate:
                v = cls._get_json_value(value) if s_type == 'JSON' else value
                cls.validate(
                    s_type,
                    v,
                    s_def.get('options'),
                    project=project,
                    user=user,
                )
            s_mod = (
                bool(s_def['user_modifiable'])
                if 'user_modifiable' in s_def
                else True
            )
            s_vals = {
                'app_plugin': app_plugin_model,
                'project': project,
                'user': user,
                'name': setting_name,
                'type': s_type,
                'user_modifiable': s_mod,
            }
            if s_type == 'JSON':
                s_vals['value_json'] = cls._get_json_value(value)
            else:
                s_vals['value'] = value
            AppSetting.objects.create(**s_vals)
            cls._log_set_debug(
                'Create', app_name, setting_name, value, project, user
            )
            return True

    @classmethod
    def delete(cls, app_name, setting_name, project=None, user=None):
        """
        Delete one or more app setting objects. In case of a PROJECT_USER
        setting, can be used to delete all settings related to project.

        :param app_name: App name (string, must equal "name" in app plugin)
        :param setting_name: Setting name (string)
        :param project: Project object to delete setting from (optional)
        :param user: User object to delete setting from (optional)
        :raise: ValueError with invalid project/user args
        """
        setting_def = cls.get_definition(setting_name, app_name=app_name)
        if setting_def['scope'] != APP_SETTING_SCOPE_PROJECT_USER:
            cls._check_project_and_user(
                setting_def.get('scope', None), project, user
            )
        elif not project:
            raise ValueError(
                'Project must be set for {} scope settings'.format(
                    APP_SETTING_SCOPE_PROJECT_USER
                )
            )
        q_kwargs = {'name': setting_name}
        if user:
            q_kwargs['user'] = user
        if project:
            q_kwargs['project'] = project
        logger.debug(
            'Delete app setting: {}.{} ({})'.format(
                app_name, setting_name, '; '.join(q_kwargs)
            )
        )
        app_settings = AppSetting.objects.filter(**q_kwargs)
        s_count = app_settings.count()
        app_settings.delete()
        logger.debug(
            'Deleted {} app setting{}'.format(
                s_count, 's' if s_count != 1 else ''
            )
        )

    @classmethod
    def delete_by_scope(
        cls,
        scope,
        project=None,
        user=None,
    ):
        """
        Delete all app settings within a given scope for a project and/or user.

        :param scope: Setting scope (string)
        :param project: Project object to delete setting from
        :param user: User object to delete setting from
        :raise: ValueError if scope, project or user are incorrect
        """
        if not scope:
            raise ValueError('Scope must be set')
        cls._check_scope(scope)
        cls._check_project_and_user(scope, project, user)
        for app_name, app_settings in cls.get_all_defs().items():
            for setting_name, setting_def in app_settings.items():
                if setting_def['scope'] == scope:
                    cls.delete(
                        app_name,
                        setting_name,
                        project=project,
                        user=user,
                    )

    @classmethod
    def validate(
        cls,
        setting_type,
        setting_value,
        setting_options,
        project=None,
        user=None,
    ):
        """
        Validate setting value according to its type.

        :param setting_type: Setting type
        :param setting_value: Setting value
        :param setting_options: Setting options (can be None)
        :param project: Project object (optional)
        :param user: User object (optional)
        :raise: ValueError if setting_type or setting_value is invalid
        """
        cls._check_type(setting_type)
        cls._check_type_options(setting_type, setting_options)
        cls._check_value_in_options(
            setting_value, setting_options, project=project, user=user
        )

        # Check callable
        if callable(setting_value):
            setting_value(project, user)

        if setting_type == 'BOOLEAN':
            if not isinstance(setting_value, bool):
                raise ValueError(
                    'Please enter a valid boolean value ({})'.format(
                        setting_value
                    )
                )
        elif setting_type == 'INTEGER':
            if (
                not isinstance(setting_value, int)
                and not str(setting_value).isdigit()
            ):
                raise ValueError(
                    'Please enter a valid integer value ({})'.format(
                        setting_value
                    )
                )
        elif setting_type == 'JSON':
            try:
                json.dumps(setting_value)
            except TypeError:
                raise ValueError(
                    'Please enter valid JSON ({})'.format(setting_value)
                )
        return True

    @classmethod
    def get_definition(cls, name, plugin=None, app_name=None):
        """
        Return definition for a single app setting, either based on an app name
        or the plugin object.

        :param name: Setting name
        :param plugin: Plugin object or None
        :param app_name: Name of the app plugin (string or None)
        :return: Dict
        :raise: ValueError if neither app_name nor plugin are set or if setting
                is not found in plugin
        """
        defs = cls._get_defs(plugin, app_name)
        if name not in defs:
            raise ValueError(
                'App setting not found in app "{}" with name "{}"'.format(
                    app_name or plugin.name, name
                )
            )
        ret = defs[name]
        cls._check_type(ret['type'])
        cls._check_type_options(ret['type'], ret.get('options'))
        return ret

    @classmethod
    def get_definitions(
        cls,
        scope,
        plugin=None,
        app_name=None,
        user_modifiable=False,
    ):
        """
        Return app setting definitions of a specific scope from a plugin.

        :param scope: PROJECT, USER or PROJECT_USER
        :param plugin: Plugin object or None
        :param app_name: Name of the app plugin (string or None)
        :param user_modifiable: Only return non-superuser modifiable settings if
                                True (boolean)
        :return: Dict
        :raise: ValueError if scope is invalid or if neither app_name nor
                plugin are set
        """
        cls._check_scope(scope)
        defs = cls._get_defs(plugin, app_name)
        ret = {
            k: v
            for k, v in defs.items()
            if (
                'scope' in v
                and v['scope'] == scope
                and (
                    not user_modifiable
                    or (
                        'user_modifiable' not in v
                        or v['user_modifiable'] is True
                    )
                )
            )
        }
        # Ensure type validity
        for k, v in ret.items():
            cls._check_type(v['type'])
            cls._check_type_options(v['type'], v.get('options'))
        return ret

    @classmethod
    def get_projectroles_defs(cls):
        """
        Return projectroles settings definitions. If it exists, get value from
        settings.PROJECTROLES_APP_SETTINGS_TEST for testing modifications.

        :return: Dict
        """
        try:
            app_settings = (
                settings.PROJECTROLES_APP_SETTINGS_TEST
                or PROJECTROLES_APP_SETTINGS
            )
        except AttributeError:
            app_settings = PROJECTROLES_APP_SETTINGS
        for k, v in app_settings.items():
            if 'local' not in v:
                raise ValueError(
                    'Attribute "local" is missing in projectroles app '
                    'setting definition "{}"'.format(k)
                )
        return app_settings

    @classmethod
    def get_all_defs(cls):
        """
        Return app setting definitions for projectroles and all active app
        plugins in a dictionary with the app name as key.

        :return: Dict
        """
        ret = {'projectroles': cls.get_projectroles_defs()}
        plugins = (
            []
            + get_active_plugins('project_app')
            + get_active_plugins('site_app')
        )
        for p in plugins:
            ret[p.name] = p.app_settings
        return ret


def get_example_setting_default(project=None, user=None):
    """
    Example method for callable default value retrieval for app settings.

    :param project: Project object
    :param user: User object
    :return: String with project and user info or 'No project'
    """
    response = 'N/A'
    if project and user:
        response = '{}:{}'.format(project.title, user.username)
    elif project:
        response = str(project.sodar_uuid)
    elif user:
        response = str(user.sodar_uuid)
    return response


def get_example_setting_options(project=None, user=None):
    """
    Example method for callable option list retrieval for app settings.

    :param project: Project object
    :param user: User object
    :return: List of tuples for ChoiceField
    """
    response = [
        ('N/A', 'No project or user for callable'),
        'Example string option',
    ]
    if project and user:
        response.append(
            (
                str(project.sodar_uuid),
                'Project UUID {} by {}'.format(
                    project.sodar_uuid, user.username
                ),
            )
        )
    elif project:
        response.append(
            (
                str(project.sodar_uuid),
                'Project UUID: {}'.format(project.sodar_uuid),
            )
        )
    elif user:
        response.append(
            (str(user.sodar_uuid), 'User UUID: {}'.format(user.sodar_uuid))
        )
    return response



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-10-10 15:34
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0005_update_uuid'),
    ]

    operations = [
        migrations.CreateModel(
            name='RemoteProject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('project_uuid', models.UUIDField(default=None, help_text='Project UUID')),
                ('level', models.CharField(default='NONE', help_text='Project access level', max_length=255)),
                ('date_access', models.DateTimeField(help_text='DateTime of last access from/to remote site', null=True)),
                ('sodar_uuid', models.UUIDField(default=uuid.uuid4, help_text='RemoteProject relation UUID (local)', unique=True)),
            ],
            options={
                'ordering': ['site__name', 'project_uuid'],
            },
        ),
        migrations.CreateModel(
            name='RemoteSite',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Site name', max_length=255, unique=True)),
                ('url', models.URLField(help_text='Site URL', max_length=2000)),
                ('mode', models.CharField(default='TARGET', help_text='Site mode', max_length=64)),
                ('description', models.TextField(help_text='Site description')),
                ('secret', models.CharField(help_text='Secret token for connecting to the source site', max_length=255)),
                ('sodar_uuid', models.UUIDField(default=uuid.uuid4, help_text='RemoteSite relation UUID (local)', unique=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.AlterField(
            model_name='project',
            name='description',
            field=models.CharField(blank=True, help_text='Short project description', max_length=512, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='remotesite',
            unique_together=set([('url', 'mode', 'secret')]),
        ),
        migrations.AddField(
            model_name='remoteproject',
            name='project',
            field=models.ForeignKey(blank=True, help_text='Related project object (if created locally)', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='remotes', to='projectroles.Project'),
        ),
        migrations.AddField(
            model_name='remoteproject',
            name='site',
            field=models.ForeignKey(help_text='Remote SODAR site', on_delete=django.db.models.deletion.CASCADE, related_name='projects', to='projectroles.RemoteSite'),
        ),
    ]



# Generated by Django 3.1.7 on 2021-04-07 15:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0019_project_public_guest_access'),
        ('timeline', '0006_update_user_optional'),
    ]

    operations = [
        migrations.AlterField(
            model_name='projectevent',
            name='project',
            field=models.ForeignKey(help_text='Project to which the event belongs (null for no project)', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='events', to='projectroles.project'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.25 on 2019-10-02 14:29
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('projectroles', '0012_update_remotesite'),
    ]

    operations = [
        migrations.AlterField(
            model_name='appsetting',
            name='type',
            field=models.CharField(help_text='Type of the setting', max_length=64),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-08-02 13:58
from __future__ import unicode_literals

from django.db import migrations, models
import markupfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('adminalerts', '0002_adminalert_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='adminalert',
            name='_description_rendered',
            field=models.TextField(editable=False, null=True),
        ),
        migrations.AddField(
            model_name='adminalert',
            name='description_markup_type',
            field=models.CharField(choices=[('', '--'), ('html', 'HTML'), ('plain', 'Plain'), ('markdown', 'Markdown'), ('restructuredtext', 'Restructured Text')], default='markdown', editable=False, max_length=30),
        ),
        migrations.AlterField(
            model_name='adminalert',
            name='description',
            field=markupfield.fields.MarkupField(blank=True, help_text='Full description of alert (optional, will be shown on a separate page)', null=True, rendered_field=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-04-11 15:58
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('projectroles', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('app', models.CharField(help_text='App from which the event was triggered', max_length=255)),
                ('event_name', models.CharField(help_text='Event ID string', max_length=255)),
                ('description', models.TextField(help_text='Description of status change (may include {object label} references)')),
                ('extra_data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, help_text='Additional event data as JSON')),
                ('classified', models.BooleanField(default=False, help_text='Event is classified (only viewable by user levels specified in rules)')),
                ('omics_uuid', models.UUIDField(default=uuid.uuid4, help_text='Event omics UUID', unique=True)),
                ('project', models.ForeignKey(help_text='Project in which the event belongs', on_delete=django.db.models.deletion.CASCADE, related_name='events', to='projectroles.Project')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectEventObjectRef',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(help_text='Label for the object related to the event', max_length=255)),
                ('name', models.CharField(help_text='Name or title of the object', max_length=255)),
                ('object_model', models.CharField(help_text='Object model as string', max_length=255)),
                ('object_uuid', models.UUIDField(blank=True, help_text='Object Omics UUID', null=True)),
                ('extra_data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, help_text='Additional data related to the object as JSON')),
                ('event', models.ForeignKey(help_text='Event to which the object belongs', on_delete=django.db.models.deletion.CASCADE, related_name='event_objects', to='timeline.ProjectEvent')),
            ],
        ),
        migrations.CreateModel(
            name='ProjectEventStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField(auto_now_add=True, help_text='DateTime of the status change')),
                ('status_type', models.CharField(help_text='Type of the status change', max_length=64)),
                ('description', models.TextField(blank=True, help_text='Description of status change (optional)')),
                ('extra_data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, help_text='Additional status data as JSON')),
                ('event', models.ForeignKey(help_text='Event to which the status change belongs', on_delete=django.db.models.deletion.CASCADE, related_name='status_changes', to='timeline.ProjectEvent')),
            ],
        ),
    ]



"""UI views for the userprofile app"""

from django.contrib import auth, messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse_lazy
from django.views.generic import TemplateView, FormView

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import get_active_plugins
from projectroles.views import (
    LoggedInPermissionMixin,
    HTTPRefererMixin,
    InvalidFormMixin,
)

from userprofile.forms import UserSettingsForm


User = auth.get_user_model()
app_settings = AppSettingAPI()


# SODAR Constants
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']

# Local Constants
SETTING_UPDATE_MSG = 'User settings updated.'


class UserDetailView(LoginRequiredMixin, LoggedInPermissionMixin, TemplateView):
    """Display the user profile view including the user settings"""

    template_name = 'userprofile/detail.html'
    permission_required = 'userprofile.view_detail'

    def _get_user_settings(self):
        plugins = get_active_plugins(
            plugin_type='project_app'
        ) + get_active_plugins(plugin_type='site_app')
        for plugin in plugins + [None]:
            if plugin:
                name = plugin.name
                p_settings = app_settings.get_definitions(
                    APP_SETTING_SCOPE_USER, plugin=plugin, user_modifiable=True
                )
            else:
                name = 'projectroles'
                p_settings = app_settings.get_definitions(
                    APP_SETTING_SCOPE_USER, app_name=name, user_modifiable=True
                )
            for k, v in p_settings.items():
                yield {
                    'label': v.get('label') or '{}.{}'.format(name, k),
                    'value': app_settings.get(name, k, user=self.request.user),
                    'description': v.get('description'),
                }

    def get_context_data(self, **kwargs):
        result = super().get_context_data(**kwargs)
        result['user_settings'] = list(self._get_user_settings())
        result['local_user'] = self.request.user.is_local()
        return result


class UserSettingsView(
    LoginRequiredMixin,
    LoggedInPermissionMixin,
    HTTPRefererMixin,
    InvalidFormMixin,
    FormView,
):
    """User settings update view"""

    form_class = UserSettingsForm
    permission_required = 'userprofile.view_detail'
    template_name = 'userprofile/settings_form.html'
    success_url = reverse_lazy('userprofile:detail')

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['current_user'] = self.request.user
        return kwargs

    def form_valid(self, form):
        result = super().form_valid(form)
        for k, v in form.cleaned_data.items():
            if k.startswith('settings.'):
                _, app_name, setting_name = k.split('.', 3)
                # TODO: Omit global USER settings (#1329)
                app_settings.set(
                    app_name, setting_name, v, user=self.request.user
                )
        messages.success(self.request, SETTING_UPDATE_MSG)
        return result
